"""
Iterate or find items in a hierarchical data structure (e.g. loaded from JSON content) based
on JSON Path-like Pythonic expressions.
"""

from .core import JP, find_all, find_next, parse
from .utilities import find_all_duplicate, find_all_unique, with_values



"""
Utility functions built on top of JSON-Path like expressions
and find_all() / find_next().
"""

# pylint: disable=unnecessary-lambda-assignment,protected-access

from copy import deepcopy
from typing import Any, Callable, Iterator

from aweson.core import _Accessor, find_all


def _find_all_with_multiplicity(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Workhorse utility function for finding both unique and duplicate items.
    """
    item_func = (lambda i: i[1]) if with_path else (lambda i: i)

    known_items: dict = {}

    for tup in find_all(root_data, path, with_path=with_path, lenient=lenient):
        item = item_func(tup)
        if item in known_items:
            known_items[item] += 1
        else:
            known_items[item] = 0
        yield tup, known_items[item]


def find_all_unique(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Yields unique elemnents, with or without paths.
    """
    yield from (
        item
        for item, multiplicity in _find_all_with_multiplicity(
            root_data, path, with_path=with_path, lenient=lenient
        )
        if multiplicity == 0
    )


def find_all_duplicate(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Yields duplicate elemnents, with or without paths.
    """
    yield from (
        item
        for item, multiplicity in _find_all_with_multiplicity(
            root_data, path, with_path=with_path, lenient=lenient
        )
        if multiplicity > 0
    )


def with_values(
    content: list | dict,
    path: _Accessor,
    values: Any | Iterator[Any] | Callable[[_Accessor, Any], Any] = None,
    /,
):
    """
    Returns a content (list or dictionary) with values specified by the call.

    The place(s) where values of the returned content differ to the original one
    is specified by ``path``. New values can come from an iterator, or they can
    be calculated, or they can be defined as values.

    The original argument ``content`` is treated immutable, i.e. the returned
    data will be a different object.

    For dictionaries, existing keys can have new values, and entirely new
    key/value pairs can be added.

    For lists, existing indexes can have new values defined, and only a single
    new item can be appended to a list.
    """
    content = deepcopy(content)

    all_accessors = list(path._accessors())
    stack = [(content, all_accessors)]

    while len(stack) > 0:
        data, accessors = stack.pop()
        if len(accessors) == 0:
            continue
        accessor = accessors[0]
        if len(accessors) == 1:
            if hasattr(values, "__iter__") and hasattr(values, "__next__"):
                iterator = values

                def insert_fun(_):
                    try:
                        return next(iterator)
                    except StopIteration as exc:
                        raise ValueError(
                            "Iterator size falls short of expectations"
                        ) from exc

            elif hasattr(values, "__call__"):
                insert_fun = values  # type: ignore
            else:
                single_value = values
                insert_fun = lambda x: single_value
            is_penultimate = True
        else:
            insert_fun = lambda x: accessors[1].container_type()
            is_penultimate = False

        stack_insert_position = len(stack)
        sub_tuples = accessor._access_or_insert(data, is_penultimate, insert_fun)
        for sub_data in sub_tuples:
            stack.insert(stack_insert_position, (sub_data, accessors[1:]))

    return content



"""
Infra for JSON Path-like expressions and finding items in data hiearchy.
"""

# pylint: disable=protected-access
from __future__ import annotations

import dataclasses as dc
import json
import re
from abc import ABC, abstractmethod
from json import JSONDecodeError
from typing import Any, Callable, Iterator


@dc.dataclass(frozen=True, kw_only=True)
class _Predicate(ABC):
    """
    Abstract base class for predicate based list-item selection.
    """

    @abstractmethod
    def _evaluate(self, content) -> bool:
        """
        Evaluates this predicate within the context of the given content.
        """


@dc.dataclass(frozen=True, kw_only=True, eq=True)
class _BinaryPredicate(_Predicate):
    """
    Binary predicate for predicate based list-item selection.
    """

    operand1: _Accessor
    operand2: Any
    func: Callable[[Any, Any], bool] = dc.field(compare=False)
    repr_template: str  # format string referring to '{op1}' and '{op2}' variables

    def _evaluate(self, content) -> bool:
        non_existent = (1,)
        operand1 = find_next(content, self.operand1, default=non_existent)
        operand2 = (
            find_next(content, self.operand2, default=non_existent)
            if isinstance(self.operand2, _Accessor)
            else self.operand2
        )
        if (operand1 is non_existent) or (operand2 is non_existent):
            return False
        if operand1 is None and operand2 is None:
            return True
        if operand1 is None or operand2 is None:
            return False
        return self.func(operand1, operand2)

    def __bool__(self):
        """
        BinaryPredicate is primarily to serve as an infra for expressions like
        ``JP[JP.field1 == JP.field2]``. But what if someone just wants to test for
        path equivalence, like ``JP.field1 == JP.field2`` or ``JP.field1 != JP.field2``?
        Then this converter dunder method kicks in.
        """
        if "==" in self.repr_template:
            return str(self.operand1) == str(self.operand2)
        if "!=" in self.repr_template:
            return str(self.operand1) != str(self.operand2)
        raise NotImplementedError("Unsupported comparison of paths")

    def __str__(self) -> str:
        op1 = f"?{self.operand1._json_path_like(child_context=True)}"
        if isinstance(self.operand2, _Accessor):
            op2 = f"?{self.operand2._json_path_like(child_context=True)}"
        else:
            op2 = json.dumps(self.operand2)
        return self.repr_template.format(op1=op1, op2=op2)


_BINARY_PREDICATE_ARGS = {
    "==": {
        "func": (lambda x, y: x == y),
        "repr_template": "{op1} == {op2}",
    },
    "!=": {
        "func": (lambda x, y: x != y),
        "repr_template": "{op1} != {op2}",
    },
    "<": {
        "func": (lambda x, y: x < y),
        "repr_template": "{op1} < {op2}",
    },
    "<=": {
        "func": (lambda x, y: x <= y),
        "repr_template": "{op1} <= {op2}",
    },
    ">": {
        "func": (lambda x, y: x > y),
        "repr_template": "{op1} > {op2}",
    },
    ">=": {
        "func": (lambda x, y: x >= y),
        "repr_template": "{op1} >= {op2}",
    },
}


@dc.dataclass(frozen=True, kw_only=True)
class _PathExistsPredicate(_Predicate):
    """
    A unary predicate telling if a sub-path exists, for predicate based list-item selection.
    """

    path: _Accessor

    def _evaluate(self, content) -> bool:
        non_existent = (1,)

        found = find_next(content, self.path, default=non_existent)
        return found is not non_existent

    def __str__(self):
        return f"?{self.path._json_path_like(child_context=True)}"


@dc.dataclass(frozen=True, kw_only=True)
class _Accessor:
    """
    Base class for building JSON Path-like expression.
    """

    parent: _Accessor | None
    container_type: type

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        """
        Args:
            container: the data structure to access into
            yield_path: whether to yield a singular path to the item (or items) that are being accessed
            lenient: whether to allow out-of-bounds indexing or missing dict key references to raise
                IndexError or KeyError, respectively

        Returns:
            An iterator to tuples, where there the first field of the tuple is the item being accessed
            at this point, and the second field of the tuple is either ``None``, or a function taking
            a single ``parent`` argument, to create paths leading to those items being accessed.
        """
        raise NotImplementedError("Should not be invoked")

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        """
        Similar to ``_access()``, except this one is used for in-place modification of
        data.

        Args:
            container: the structure (list or dict) being accessed (or inserted into)
            is_penultimate: whether this structure (list or dict) is the last-but-one
            segment in the JSON Path-like expression, i.e. the immediate parent of
            a leaf data being inserted into.
            insert_fun: a unary function providing the new value to insert, the argument
                is the current value if exists, or ``None`` otherwise

        Returns:
            An iterator of child structures to further descend into with the JSON Path
            if is_penultimate=False, or the (newly) inserted leaf values if
            is_penultimate=True.
        """
        raise NotImplementedError("Should not be invoked")

    def _is_singular(self) -> bool:
        """
        Returns: if this accessor (excluding any parent from consideration) can only
            ever return a single item.
        """
        return True

    def is_singular(self) -> bool:
        """
        Returns: if this path (this accessor and its parents, transitively) can only
            ever return a single item. E.g. a slice expression in a path makes it non-singular,
            even if the start/stop combination otherwise would say it's singular, like ``[1:2]``.
        """
        return self._is_singular() and (
            self.parent is None or self.parent.is_singular()
        )

    def _representation(self) -> str:
        """
        Represention for this accessor (excluding any parent)
        """
        raise NotImplementedError("Root accessor should not be invoked")

    def _check_container_type(self, container: list | dict):
        """
        Performs a check that a container to access is of the type mandated by
        the declared container_type of an accessor instance.
        """
        if not isinstance(  # pylint: disable=isinstance-second-argument-not-valid-type
            container, self.container_type
        ):
            raise ValueError(
                f"Expected {self.container_type}, got {type(container)} at {self}"
            )

    def _accessors(self) -> list[_Accessor]:
        """
        List of accessors, from root to this, recursively collected.
        """
        if self.parent is None:
            # Symmetry would suggest here to `return [self]`, however,
            # we cheat here: this instance shall be the root (=first accessor
            # to traverse by), not the parent
            return []
        return self.parent._accessors() + [self]

    def __str__(self):
        return self._json_path_like()

    def _json_path_like(self, child_context: bool = False):
        """
        A (best effort attempt) to render the path, made up
        by this accessor and its parents, transitively, as a JSON Path.

        Args:
            child_context: JSON Path uses marker "$" for paths starting
            from document root, and "@" for paths starting from a child
            node. This flag controls which context to build the string for.
        """
        accessors = self._accessors()
        marker = "@" if child_context else "$"
        return marker + "".join(a._representation() for a in accessors)

    def __getattr__(self, specification):
        """
        JSON Path-like expression builder infra.

        Overloaded for dict key access.
        """
        return _DictKeyAccessor(parent=self, key=specification)

    def __getitem__(self, specification):  # pylint: disable=too-many-return-statements
        """
        JSON Path-like expression builder infra.

        Overloaded for list (index, slice) and various dict key access.
        """
        if isinstance(specification, str):
            if specification == "*":
                return _ListSliceAccessor(parent=self, slice_=slice(None, None, None))
            if specification.isidentifier():
                return _DictKeyAccessor(parent=self, key=specification)
            key_regex = re.compile(specification)
            return _DictKeyRegexAccessor(parent=self, key_regex=key_regex)
        if isinstance(specification, _Accessor):
            return _ListPredicateAccessor(
                parent=self, predicate=_PathExistsPredicate(path=specification)
            )
        if isinstance(specification, _Predicate):
            return _ListPredicateAccessor(parent=self, predicate=specification)
        if isinstance(specification, int):
            return _ListIndexAccessor(parent=self, index=specification)
        if isinstance(specification, slice):
            return _ListSliceAccessor(parent=self, slice_=specification)
        raise ValueError(f"Unsupported indexing expression {specification}")

    def __call__(self, *paths, **named_paths):
        """
        JSON Path-like expression builder infra.

        Overloaded for sub-item selection (vanilla or named tuple).
        """
        if len(paths) > 0 and len(named_paths) > 0:
            raise NotImplementedError(
                "Either all sub-selections are to be named, or none of them."
            )

        def verify_paths(paths):
            not_accessors = [path for path in paths if not isinstance(path, _Accessor)]
            if len(not_accessors) > 0:
                raise ValueError(f"Not paths: {not_accessors}")
            not_singulars = [path for path in paths if not path.is_singular()]
            if len(not_singulars) > 0:
                raise ValueError(
                    f"Not singular paths (could point to multiple items): {not_singulars}"
                )

        if len(paths) > 0:
            verify_paths(paths)
            return _SubHiearchyAccessor(
                parent=self,
                sub_accessors=paths,
                sub_hierarchy_ctor=tuple,
                field_name_mapping=tuple(),
            )
        if len(named_paths) > 0:
            paths = list(named_paths.values())
            verify_paths(paths)
            return _SubHiearchyAccessor(
                parent=self,
                sub_accessors=paths,
                sub_hierarchy_ctor=lambda values: dict(zip(named_paths.keys(), values)),
                field_name_mapping=tuple(named_paths.keys()),
            )
        raise NotImplementedError("Sub-selection cannot be empty")

    def __eq__(self, other):
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["=="],
        )

    def __ne__(self, other):
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["!="],
        )

    def __gt__(self, other):
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS[">"],
        )

    def __ge__(self, other):
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS[">="],
        )

    def __lt__(self, other):
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["<"],
        )

    def __le__(self, other):
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["<="],
        )


@dc.dataclass(frozen=True, kw_only=True)
class _DictKeyAccessor(_Accessor):
    """Accesses a value of a dict container by a key"""

    key: str
    container_type: type = dict

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        self._check_container_type(container)
        if lenient and self.key not in container:
            yield from iter([])
        elif yield_path:
            yield container[self.key], (lambda parent: _DictKeyAccessor(parent=parent, key=self.key))  # type: ignore
        else:
            yield container[self.key], None  # type: ignore

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        self._check_container_type(container)
        if self.key not in container or is_penultimate:
            container[self.key] = insert_fun(container.get(self.key, None))  # type: ignore

        yield container[self.key]  # type: ignore

    def _representation(self) -> str:
        return f".{self.key}"

    def __eq__(self, other):
        """
        Overloaded == operator in the superclass does not work in sub-classes.
        Other operators don't seem to have any trouble.
        """
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["=="],
        )


@dc.dataclass(frozen=True, kw_only=True)
class _DictKeyRegexAccessor(_Accessor):
    """Accesses a value or values of a dict container by a regex matching keys"""

    key_regex: re.Pattern
    container_type: type = dict

    def _is_singular(self) -> bool:
        return False

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        self._check_container_type(container)
        for key, value in container.items():  # type: ignore
            if self.key_regex.findall(key):  # pylint: disable=no-member
                if yield_path:
                    yield value, lambda parent: _DictKeyAccessor(
                        parent=parent, key=key  # pylint: disable=cell-var-from-loop
                    )
                else:
                    yield value, None

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        self._check_container_type(container)
        for key in container.keys():  # type: ignore
            if self.key_regex.findall(key):  # pylint: disable=no-member
                if is_penultimate:
                    container[key] = insert_fun(container[key])
                yield container[key]

    def _representation(self) -> str:
        return f"[{self.key_regex.pattern}]"  # pylint: disable=no-member


@dc.dataclass(frozen=True, kw_only=True)
class _ListIndexAccessor(_Accessor):
    """Accesses an item of a list by an index"""

    index: int
    container_type: type = list

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        self._check_container_type(container)
        if lenient and (self.index >= len(container) or self.index < -len(container)):
            yield from iter([])
        elif yield_path:
            if self.index >= 0:
                yield container[self.index], lambda parent: _ListIndexAccessor(
                    parent=parent, index=self.index
                )
            else:
                yield container[self.index], lambda parent: _ListIndexAccessor(
                    parent=parent, index=len(container) + self.index
                )
        else:
            yield container[self.index], None

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        self._check_container_type(container)
        if self.index == len(container):
            container.append(insert_fun(None))  # type: ignore
        elif self.index < len(container):
            if is_penultimate:
                container[self.index] = insert_fun(container[self.index])
            else:
                pass
        else:
            raise ValueError(
                f"Cannot append at {self.index} to a list with length {len(container)}"
            )
        yield container[self.index]

    def _representation(self) -> str:
        return f"[{self.index}]"

    def __eq__(self, other):
        """
        Overloaded == operator in the superclass does not work in sub-classes.
        Other operators don't seem to have any trouble.
        """
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["=="],
        )


def _create_list_idx_accessor_ctor(index: int) -> _Accessor:
    return lambda parent: _ListIndexAccessor(parent=parent, index=index)  # type: ignore


@dc.dataclass(frozen=True, kw_only=True)
class _ListPredicateAccessor(_Accessor):
    """Accesses items of a list by a predicate"""

    predicate: _Predicate
    container_type: type = list

    def _is_singular(self) -> bool:
        return False

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        _ = lenient  # unused
        self._check_container_type(container)
        if yield_path:

            yield from (
                (item, _create_list_idx_accessor_ctor(current_index))
                for current_index, item in enumerate(container)
                if self.predicate._evaluate(item)
            )
        else:
            yield from (
                (item, None) for item in container if self.predicate._evaluate(item)
            )

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        self._check_container_type(container)
        if is_penultimate:
            for index, item in enumerate(container):
                if self.predicate._evaluate(item):
                    container[index] = insert_fun(container[index])
                    yield container[index]
        else:
            yield from (item for item in container if self.predicate._evaluate(item))

    def _representation(self) -> str:
        return f"[{self.predicate}]"

    def __eq__(self, other):
        """
        Overloaded == operator in the superclass does not work in sub-classes.
        Other operators don't seem to have any trouble.
        """
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["=="],
        )


@dc.dataclass(frozen=True, kw_only=True)
class _ListSliceAccessor(_Accessor):
    """Accesses items of a list by a slice"""

    slice_: slice
    container_type: type = list

    def _is_singular(self) -> bool:
        return False

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        _ = lenient  # unused
        self._check_container_type(container)
        if yield_path:
            slice_indices = self.slice_.indices(len(container))

            yield from (
                (item, _create_list_idx_accessor_ctor(current_index))
                for current_index, item in zip(
                    range(slice_indices[0], slice_indices[1], slice_indices[2]),
                    container[self.slice_],
                )
            )
        else:
            yield from ((item, None) for item in container[self.slice_])

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        self._check_container_type(container)
        if is_penultimate:
            slice_indices = self.slice_.indices(len(container))
            for index in range(slice_indices[0], slice_indices[1], slice_indices[2]):
                container[index] = insert_fun(container[index])
                yield container[index]
        else:
            yield from container[self.slice_]

    def _representation(self) -> str:
        repr_ = (
            (f"[{self.slice_.start}" if self.slice_.start is not None else "[")
            + (f":{self.slice_.stop}" if self.slice_.stop is not None else ":")
            + (f":{self.slice_.step}]" if self.slice_.step is not None else "]")
        )
        return repr_

    def __eq__(self, other):
        """
        Overloaded == operator in the superclass does not work in sub-classes.
        Other operators don't seem to have any trouble.
        """
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["=="],
        )


@dc.dataclass(frozen=True, kw_only=True)
class _SubHiearchyAccessor(_Accessor):
    """
    Instead of returning an entire item (of a list), it constructs a tuple based on sub-JSON Path-like expressions.
    """

    sub_accessors: list[_Accessor]
    sub_hierarchy_ctor: Callable
    container_type: type = dict
    field_name_mapping: tuple[str, ...]

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        _ = lenient  # unused
        self._check_container_type(container)
        items = [
            find_next(container, sub_accessor, default=None)
            for sub_accessor in self.sub_accessors
        ]
        if yield_path:
            yield self.sub_hierarchy_ctor(items), lambda parent: _SubHiearchyAccessor(
                parent=parent,
                sub_accessors=self.sub_accessors,
                sub_hierarchy_ctor=self.sub_hierarchy_ctor,
                field_name_mapping=self.field_name_mapping,
            )
        else:
            yield self.sub_hierarchy_ctor(items), None

    def _access_or_insert(
        self, container: list | dict, is_penultimate: bool, insert_fun: Callable
    ):
        self._check_container_type(container)
        for accessor in self.sub_accessors:
            if (
                not isinstance(accessor, _DictKeyAccessor)
                or len(accessor._accessors()) > 2
            ):
                raise ValueError(f"Unsupported mutator in sub-hierarchy: {accessor}")

        values = insert_fun(container)
        if len(values) != len(self.sub_accessors):
            raise ValueError(
                f"Value count ({len(values)}) != specs count ({len(self.sub_accessors)})"
            )

        for sub_accessor, value in zip(self.sub_accessors, values):
            if sub_accessor.key not in container or is_penultimate:
                container[sub_accessor.key] = value

        yield values

    def _representation(self):
        sub_paths = [
            sub_acc._json_path_like(child_context=True)
            for sub_acc in self.sub_accessors
        ]
        if self.field_name_mapping:
            assert len(self.field_name_mapping) == len(self.sub_accessors)
            expression = ", ".join(
                (
                    f"{name}={sub_path}"
                    for name, sub_path in zip(self.field_name_mapping, sub_paths)
                )
            )
        else:
            expression = ", ".join(sub_paths)
        # return f"({', '.join(sub_paths)})"
        return f"({expression})"

    def __eq__(self, other):
        """
        Overloaded == operator in the superclass does not work in sub-classes.
        Other operators don't seem to have any trouble.
        """
        return _BinaryPredicate(
            operand1=self,
            operand2=other,
            **_BINARY_PREDICATE_ARGS["=="],
        )


JP = _Accessor(parent=None, container_type=type(None))


def _parse_operand(operand: str) -> None | str | int | float | bool | _Accessor:
    if operand.startswith("?@"):
        path_maybe = operand.replace("?@", "$", 1)
        return parse(path_maybe)

    try:
        value = json.loads(operand)
        return value
    except JSONDecodeError as err:
        raise ValueError(f"{operand} is not a JSON value") from err


PATTERN_SEGMENT = re.compile(
    """
        \\[(?P<slice1_stop>-?\\d+)]
        | \\[(?P<slice2>(?P<slice2_start>-?\\d+)?:(?P<slice2_stop>-?\\d+)?)]
        | \\[(?P<star>\\*)]
        | \\[(?P<slice3>(?P<slice3_start>-?\\d+)?:(?P<slice3_stop>-?\\d+)?:(?P<slice3_step>-?\\d+)?)]
        | \\[(?P<tested_path>\\?@[^ ]+)]
        | \\[(?P<operand1>[^ ]+)\\ *(?P<operator><=|>=|==|!=|<|>)\\ *(?P<operand2>.+)]
        | \\.(?P<field>[^.[(]+)
        | \\[(?P<regex>.*)]
        | \\((?P<subitems>[^=]+)\\)
        | \\((?P<named_subitems>(.*=.*)+)\\)
    """,
    re.VERBOSE,
)


def parse(  # pylint: disable=too-many-locals,too-many-branches
    json_path: str,
) -> _Accessor:
    """
    Parses a stringified form of a JSON Path-like object. There is an overlap with
    standard JSON Path expressions, but there are JSON Path expressions not supported by
    this library and features of this library which are not standard JSON Path expressions.
    """
    if not json_path.startswith(("$", "@")):
        raise ValueError("TODO")

    accessor = JP

    path = json_path[1:]

    for segment_match in PATTERN_SEGMENT.finditer(path):
        if field := segment_match.group("field"):
            accessor = accessor[field]
        elif slice1_stop := segment_match.group("slice1_stop"):
            accessor = accessor[int(slice1_stop)]
        elif segment_match.group("slice2"):
            start = segment_match.group("slice2_start")
            stop = segment_match.group("slice2_stop")
            s = slice(int(start) if start else None, int(stop) if stop else None)
            accessor = accessor[s]
        elif segment_match.group("slice3"):
            start = segment_match.group("slice3_start")
            stop = segment_match.group("slice3_stop")
            step = segment_match.group("slice3_step")
            s = slice(
                int(start) if start else None,
                int(stop) if stop else None,
                int(step) if step else None,
            )
            accessor = accessor[s]
        elif tested_path := segment_match.group("tested_path"):
            path = _parse_operand(tested_path)  # type: ignore
            predicate = _PathExistsPredicate(path=path)  # type: ignore
            accessor = accessor[predicate]
        elif operator := segment_match.group("operator"):
            operand1 = _parse_operand(segment_match.group("operand1"))
            operand2 = _parse_operand(segment_match.group("operand2"))
            predicate = _BinaryPredicate(
                operand1=operand1, operand2=operand2, **_BINARY_PREDICATE_ARGS[operator]  # type: ignore
            )
            accessor = accessor[predicate]
        elif segment_match.group("star"):
            accessor = accessor[:]
        elif regex := segment_match.group("regex"):
            _ = re.compile(regex)
            accessor = accessor[regex]
        elif subitems := segment_match.group("subitems"):
            subitem_list = [parse(subitem.strip()) for subitem in subitems.split(",")]
            accessor = accessor(*subitem_list)
        elif named_subitems := segment_match.group("named_subitems"):
            key_value_pattern = re.compile("(?P<key>\\w+)\\s*=\\s*(?P<value>\\S+)")
            named_subitem_list = {
                match["key"]: parse(match["value"])
                for named_subitem in named_subitems.split(",")
                if (match := key_value_pattern.match(named_subitem.strip()))
            }
            accessor = accessor(**named_subitem_list)
        else:
            raise ValueError("TODO")

    return accessor


def find_all(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Finds all matching items in a JSON-like data hierarchy (lists of / dicts of / values) based
    on a JSON Path-like specification. Technically, it iterates over the matching items.

    Args:
        path: JSON Path-like expression, specifying what item (or items) to match & iterate over
        with_path: whether to yield accurate, JSON Path-like pointer objects to items found
        lenient: whether to allow out of bound indices or missing keys, or raise ``IndexError`` and
            ``KeyError`` exceptions, respectivately.
    """
    all_accessors = list(path._accessors())
    stack = [(root_data, all_accessors, JP if with_path else None)]

    while len(stack) > 0:
        data, accessors, current_accessor = stack.pop()
        if len(accessors) == 0:  # leaf item
            if with_path:
                yield current_accessor, data
            else:
                yield data
        else:
            accessor = accessors[0]

            # With a stack content [...] and items A, B, C iterated by accessor._access(...)
            # we want the following stack content: [..., C*, B*, A*]
            # - where A*, B*, C* are tuples created for A, B, C respectively
            # the point is that we want to process, in the next loop, in the order A*. B*, C*.
            #
            # We don't want to do an equivalent `for ... in reversed(list(accessor._access(...))):`,
            # as reversal requires constructing a full list first in order to reverse the order.
            #
            # Inserting into the Nth position (N is current length of stack) achieves the same.
            stack_insert_position = len(stack)
            sub_tuples = accessor._access(data, yield_path=with_path, lenient=lenient)  # type: ignore
            for sub_data, accessor_ctor in sub_tuples:
                new_accessor = accessor_ctor(current_accessor) if accessor_ctor is not None else None  # type: ignore
                stack.insert(
                    stack_insert_position, (sub_data, accessors[1:], new_accessor)
                )


def find_next(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    **kwargs,
):
    """
    Shorthand for ``next(find_all(...))``. Also takes a keyword argument, ``default``,
    to delegate it to the ``next(..., default=...)`` call, if defined.
    """
    if "default" in kwargs:
        default = kwargs["default"]
        try:
            # we don't want to pass the default value as `next(..., default)` ...
            return next(find_all(root_data, path, with_path=with_path, lenient=True))
        except StopIteration:
            # ... because we need to return None for path, if with_path=True
            return (None, default) if with_path else default
    else:
        return next(find_all(root_data, path, with_path=with_path))



from itertools import product

import pytest

from aweson import JP, find_all, find_next
from aweson.core import _Accessor, _Predicate, parse, _BinaryPredicate


###############################################################
#
# infra
#


def test_root_accessors_are_equivalent():
    a_root = _Accessor(parent=None, container_type=type(None))
    assert JP == a_root
    assert a_root == _Accessor(parent=None, container_type=type(None))


@pytest.mark.parametrize("path1, path2", [
    (JP, JP),
    (JP.hello, JP.hello),
    (JP.hello[2], JP.hello[2]),
    (JP.hello[2:4], JP.hello[2:4]),
    (JP.hello["hi"], JP.hello["hi"]),
    (JP.hello[JP.hi == "hi"], JP.hello[JP.hi == "hi"]),
    (JP.hello(JP.hi, JP.hola[0]), JP.hello(JP.hi, JP.hola[0])),
    (JP.hello(world=JP.mundo, hi=JP.hola[0]), JP.hello(world=JP.mundo, hi=JP.hola[0]))
])
def test_equivalent_paths(path1, path2):
    # tests equivalent to different, but equivalent path object
    assert isinstance(path1 == path2, _BinaryPredicate)
    assert path1 == path2  # truthy

    # tests equivalent to self
    assert path1 == path1


@pytest.mark.parametrize(
    "path1, path2",
    [
        (JP, JP.hello),
        (JP, JP[0]),
        (JP.hello, JP[0]),
        (JP.hello, JP.hi),
        (JP[0], JP[1]),
    ],
)
def test_different_paths_are_not_equivalent(path1, path2):
    assert path1 != path2


@pytest.mark.parametrize(
    "path1, path2",
    [
        (JP["hi"], JP.hi),
        (JP[0]["hi"], JP[0].hi),
        (JP["hi"][1], JP.hi[1]),
    ],
)
def test_field_access_is_equivalent_to_idx_by_string(path1, path2):
    assert path1 == path2


@pytest.mark.parametrize(
    "jp_bool",
    [
        JP == 0,
        JP != 0,
        JP < 0,
        JP <= 0,
        JP > 0,
        JP >= 0,
        JP.price == 0,
        JP.price != 0,
        JP.price < 0,
        JP.price <= 0,
        JP.price > 0,
        JP.price >= 0,
        JP.price[0] == 0,
        JP.price[0] != 0,
        JP.price[0] < 0,
        JP.price[0] <= 0,
        JP.price[0] > 0,
        JP.price[0] >= 0,
        JP.price == 0,
        JP.price != 0,
        JP.price < 0,
        JP.price <= 0,
        JP.price > 0,
        JP.price >= 0,
        JP.price.unit == 0,
        JP.price.unit != 0,
        JP.price.unit < 0,
        JP.price.unit <= 0,
        JP.price.unit > 0,
        JP.price.unit >= 0,
    ],
)
def test_overloaded_bool_operators(jp_bool):
    assert isinstance(jp_bool, _Predicate)


###############################################################
#
# str()
#


PATH_OBJECTS_AND_STRINGIFIED = [
    (JP, "$"),
    (JP.hello, "$.hello"),
    (JP.hello.world, "$.hello.world"),
    (JP["hello"], "$.hello"),
    (JP["hello"]["world"], "$.hello.world"),
    (JP[0], "$[0]"),
    (JP[42], "$[42]"),
    (JP[-1], "$[-1]"),
    (JP[1:2], "$[1:2]"),
    (JP[-2:-1], "$[-2:-1]"),
    (JP[1:], "$[1:]"),
    (JP[:2], "$[:2]"),
    (JP[:-1], "$[:-1]"),
    (JP[12:], "$[12:]"),
    (JP[-2:], "$[-2:]"),
    (JP[5][42], "$[5][42]"),
    (JP[:], "$[:]"),
    (JP["*"], "$[:]"),
    # (JP[:], "$[*]"),
    (JP[::], "$[:]"),
    (JP[1:], "$[1:]"),
    (JP[1::], "$[1:]"),
    (JP[:2], "$[:2]"),
    (JP[:2:], "$[:2]"),
    (JP[::-1], "$[::-1]"),
    (JP[1:12:-1], "$[1:12:-1]"),
    (JP[::-1], "$[::-1]"),
    (JP[1:12:-1], "$[1:12:-1]"),
    (JP[5].hello, "$[5].hello"),
    (JP.hello[5], "$.hello[5]"),
    (JP[5].hello[42].world, "$[5].hello[42].world"),
    (JP.hello[5].world[42], "$.hello[5].world[42]"),
    (JP["hello"][5]["world"][42], "$.hello[5].world[42]"),
    (JP[JP == 0], "$[?@ == 0]"),
    (JP[JP.id == 0], "$[?@.id == 0]"),
    (JP[JP.id == None], "$[?@.id == null]"),
    (JP[JP != 0], "$[?@ != 0]"),
    (JP[JP > 0], "$[?@ > 0]"),
    (JP[JP >= 0], "$[?@ >= 0]"),
    (JP[JP < 0], "$[?@ < 0]"),
    (JP[JP <= 0], "$[?@ <= 0]"),
    (JP[JP.id], "$[?@.id]"),
    (JP[JP.id == 0], "$[?@.id == 0]"),
    (JP.products[JP == 0], "$.products[?@ == 0]"),
    (JP.products[JP.id], "$.products[?@.id]"),
    (JP.products[JP.price > 120], "$.products[?@.price > 120]"),
    (JP.products[JP.price > 120.1], "$.products[?@.price > 120.1]"),
    (JP.products[JP.invalid == False], "$.products[?@.invalid == false]"),
    (JP.products[JP.name == "acme"], "$.products[?@.name == \"acme\"]"),
    (JP.products[JP.name == None], "$.products[?@.name == null]"),
    (JP[JP.data == [0, 1]].id, "$[?@.data == [0, 1]].id"),
    (JP[JP.data == {"first": 10, "second": 11}].id, '$[?@.data == {"first": 10, "second": 11}].id'),
    (JP.products[JP.price > JP.avg_price], "$.products[?@.price > ?@.avg_price]"),
    (JP["hello|hi"].id, "$[hello|hi].id"),
    (JP[5].hello[42].world, "$[5].hello[42].world"),
    (JP.hello[5].world[42], "$.hello[5].world[42]"),
    (JP["hello"][5]["world"][42], "$.hello[5].world[42]"),
    (JP["hello|hi"].id, "$[hello|hi].id"),
    (JP["a.*b"].id, "$[a.*b].id"),
    (JP.hello(JP.world, JP.hi[0]), "$.hello(@.world, @.hi[0])"),
    (
        JP.hello(world=JP.mundo, hi=JP.hola[0]),
        "$.hello(world=@.mundo, hi=@.hola[0])",
    ),
]


@pytest.mark.parametrize("jp,stringified", PATH_OBJECTS_AND_STRINGIFIED)
def test_jp_stringification(jp, stringified):
    assert str(jp) == stringified


###############################################################
#
# parse()
#


@pytest.mark.parametrize(
    "jp, path",
    PATH_OBJECTS_AND_STRINGIFIED + [
        (JP[:], "$[*]")
    ]
)
def test_parse(jp, path):
    parsed_jp = parse(path)
    assert parsed_jp == jp


@pytest.mark.parametrize("path",[
    "$[?@.is == (1, 2)]"
])
def test_parse_rejects(path):
    with pytest.raises(ValueError, match="not a JSON value"):
        _ = parse(path)


###############################################################
#
# is singular
#


@pytest.mark.parametrize(
    "jp, is_singular",
    [
        (JP, True),
        (JP.hello, True),
        (JP.hello[0], True),
        (JP.hello[-1], True),
        (JP["hello"], True),
        (JP[:], False),
        (JP["*"], False),
        (JP[1:2], False),
        (JP.hello[:], False),
        (JP[1:2].hello, False),
    ],
)
def test_path_is_singular(jp, is_singular):
    assert jp.is_singular() == is_singular


###############################################################
#
# sub-selection
#


def test_sub_selection_either_all_unnamed_or_all_named_fields():
    with pytest.raises(NotImplementedError):
        JP.hello(JP.some_field, named=JP.other_field)


def test_sub_selection_must_not_be_empty():
    with pytest.raises(NotImplementedError):
        JP.hello()


def test_sub_selection_must_be_path():
    with pytest.raises(ValueError):
        JP[:](JP.id, "somestring")

    with pytest.raises(ValueError):
        JP[:](id_=JP.id, detail="somestring")


def test_sub_selection_must_be_singular():
    with pytest.raises(ValueError):
        JP[:](JP.id, JP.detail[:])

    with pytest.raises(ValueError):
        JP[:](id_=JP.id, detail=JP.detail[:])


###############################################################
#
# find_all()
#


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ("string", JP, ["string"]),
        (5, JP, [5]),
        (5.13, JP, [5.13]),
        (True, JP, [True]),
        ({"hello": 42}, JP, [{"hello": 42}]),
        ([5, 42], JP, [[5, 42]]),
    ],
)
def test_find_all_for_root_path(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ({"hello": 42, "hi": "irrelevant"}, JP.hello, [42]),
        ({"hello": {"world": 42}, "hi": "irrelevant"}, JP.hello.world, [42]),
        ({"hello": 42, "hi": "irrelevant"}, JP["hello"], [42]),
        ({"hello": {"world": 42}, "hi": "irrelevant"}, JP["hello"]["world"], [42]),
    ],
)
def test_find_all_simple_key_based_dict_access(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ([5, 42], JP[0], [5]),
        ([5, 42], JP[-1], [42]),
        ([5, 42, 137], JP[:], [5, 42, 137]),
        ([5, 42, 137], JP["*"], [5, 42, 137]),
        ([5, 42, 137], JP[1:], [42, 137]),
        ([5, 42, 137], JP[:1], [5]),
        ([5, 42, 137], JP[::-1], [137, 42, 5]),
    ],
)
def test_find_all_list_access_by_index_and_slice(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            {
                "apple": [{"name": "red delicious"}, {"name": "punakaneli"}],
                "pear": [{"name": "wilhelm"}, {"name": "conference"}],
            },
            JP[".*"][:].name,
            ["red delicious", "punakaneli", "wilhelm", "conference"],
        ),
        (
            {
                "apple": [{"name": "red delicious"}, {"name": "punakaneli"}],
                "pear": [{"name": "wilhelm"}, {"name": "conference"}],
            },
            JP["apple|pear"][:].name,
            ["red delicious", "punakaneli", "wilhelm", "conference"],
        ),
    ],
)
def test_find_all_access_by_key_regex(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            {
                "apple": [{"name": "red delicious"}, {"name": "punakaneli"}],
                "pear": [{"name": "wilhelm"}, {"name": "conference"}],
            },
            JP[".*"][:].name,
            [
                (JP.apple[0].name, "red delicious"),
                (JP.apple[1].name, "punakaneli"),
                (JP.pear[0].name, "wilhelm"),
                (JP.pear[1].name, "conference"),
            ],
        ),
    ],
)
def test_find_all_access_by_key_regex_with_path(content, jp, expected_items):
    items = list(find_all(content, jp, with_path=True))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ({"hello": [5, 42]}, JP.hello[1], [42]),
        ([{"hello": 42}, {"hello": 5}], JP[1].hello, [5]),
        ([{"hello": 42}, {"hello": 5}], JP[1]["hello"], [5]),
        (
            {"l1": [{"detail": {"data": 5}}, {"detail": {"data": 42}}]},
            JP.l1[:].detail.data,
            [5, 42],
        ),
    ],
)
def test_find_all_mixed_hierarchy(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[:](JP.hello, JP.hi[1]),
            [("world", 42), ("mundo", -42)],
        ),
        (
            {
                "employees": [
                    {"name": "Doe, John", "age": 32, "account": "johndoe"},
                    {"name": "Doe, Jane", "age": -23, "account": "janedoe"},
                    {"name": "Deer, Jude", "age": 42, "account": "judedeer"},
                ]
            },
            JP.employees[:2](JP.account, JP.name),
            [("johndoe", "Doe, John"), ("janedoe", "Doe, Jane")],
        ),
    ],
)
def test_find_all_sub_selection_tuple(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[:](id=JP.hello, detail=JP.hi[1]),
            [{"id": "world", "detail": 42}, {"id": "mundo", "detail": -42}],
        ),
        (
            {
                "employees": [
                    {"name": "Doe, John", "age": 32, "account": "johndoe"},
                    {"name": "Doe, Jane", "age": -23, "account": "janedoe"},
                    {"name": "Deer, Jude", "age": 42, "account": "judedeer"},
                ]
            },
            JP.employees[:2](id=JP.account, detail=JP.name),
            [
                {"id": "johndoe", "detail": "Doe, John"},
                {"id": "janedoe", "detail": "Doe, Jane"},
            ],
        ),
    ],
)
def test_find_all_sub_selection(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


def test_find_all_demo_transformation():
    list_of_entities = [
        {"id": 5, "value": "five"},
        {"id": 42, "value": "life, universe and everything"},
        {"id": 137, "value": "137"},
    ]

    id_to_value_map = {
        id_: value for id_, value in find_all(list_of_entities, JP[:](JP.id, JP.value))
    }
    assert id_to_value_map == {
        5: "five",
        42: "life, universe and everything",
        137: "137",
    }


def test_find_all_demo_transformation_named_tuples():
    list_of_entities = [
        {"id": 5, "value": "five"},
        {"id": 42, "value": "life, universe and everything"},
        {"id": 137, "value": "137"},
    ]

    id_to_value_map = {
        tup["id"]: tup["value"]
        for tup in find_all(list_of_entities, JP[:](id=JP.id, value=JP.value))
    }
    assert id_to_value_map == {
        5: "five",
        42: "life, universe and everything",
        137: "137",
    }


@pytest.mark.parametrize(
    "content,jp,expected_paths_and_items",
    [
        ("string", JP, [(JP, "string")]),
        (
            {"hello": {"world": 42}, "hi": "irrelevant"},
            JP.hello.world,
            [(JP.hello.world, 42)],
        ),
        ([5, 42], JP[-1], [(JP[1], 42)]),
        ([5, 42, 137], JP[1:], [(JP[1], 42), (JP[2], 137)]),
        ([5, 42, 137], JP[::-1], [(JP[2], 137), (JP[1], 42), (JP[0], 5)]),
        ([{"hello": 42}, {"hello": 5}], JP[1:]["hello"], [(JP[1].hello, 5)]),
        (
            [{"hello": 42}, {"hello": 5}],
            JP[-1::-1]["hello"],
            [(JP[1].hello, 5), (JP[0].hello, 42)],
        ),
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[:](JP.hello, JP.hi[1]),
            [
                (JP[0](JP.hello, JP.hi[1]), ("world", 42)),
                (JP[1](JP.hello, JP.hi[1]), ("mundo", -42)),
            ],
        ),
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[1:0:-1](JP.hello, JP.hi[1]),
            [(JP[1](JP.hello, JP.hi[1]), ("mundo", -42))],
        ),
    ],
)
def test_find_all_with_paths(content, jp, expected_paths_and_items):
    paths_and_items = list(find_all(content, jp, with_path=True))
    assert len(paths_and_items) == len(expected_paths_and_items)
    for path_and_item, expected_path_and_item in zip(
        paths_and_items, expected_paths_and_items
    ):

        path, item = path_and_item

        # The path yielded alongside the item points to the item itself ...
        assert item == next(find_all(content, path))

        # ... but let's check it against explicit expectations, too.
        expected_path, expected_item = expected_path_and_item
        assert item == expected_item
        assert str(path) == str(expected_path)


@pytest.mark.parametrize(
    "content, jp",
    [
        ([], JP[0]),
        ([5, 42], JP[2]),
        ({"hello": [5, 42]}, JP.hello[2]),
        ({"hello": [5, 42]}, JP["hello"][2]),
    ],
)
def test_find_all_index_error(content, jp):
    with pytest.raises(IndexError):
        list(find_all(content, jp))


@pytest.mark.parametrize(
    "content, jp",
    [
        ([], JP[0]),
        ([5, 42], JP[2]),
        ([5, 42], JP[-3]),
        ({"hello": [5, 42]}, JP.hello[2]),
        ({"hello": [5, 42]}, JP["hello"][2]),
    ],
)
def test_lenient_find_all_yields_nothing_on_nonexistent_indexes(content, jp):
    assert list(find_all(content, jp, lenient=True)) == []


@pytest.mark.parametrize(
    "content, jp",
    [
        ({}, JP.hello),
        ({"hello": {"world": 42}}, JP.hello.hi),
        ({"hello": {"world": 42}}, JP["hello"]["hi"]),
    ],
)
def test_find_all_key_error(content, jp):
    with pytest.raises(KeyError):
        list(find_all(content, jp))


@pytest.mark.parametrize(
    "content, jp",
    [
        ({}, JP.hello),
        ({"hello": {"world": 42}}, JP.hello.hi),
        ({"hello": {"world": 42}}, JP["hello"]["hi"]),
    ],
)
def test_lenient_find_all_yields_nothing_on_nonexistent_keys(content, jp):
    assert list(find_all(content, jp, lenient=True)) == []


@pytest.mark.parametrize(
    "content, jp",
    [
        ([], JP[:].hello[13].hi),
        ({}, JP.hello[13].hi[:]),
    ],
)
def test_lenient_find_all_yields_nothing_mixed_case(content, jp):
    assert list(find_all(content, jp, lenient=True)) == []


@pytest.mark.parametrize(
    "content,jp,with_path,expected_items",
    [
        # cases: if attribute / path exists
        ([{"hello": 5, "id": 1}, {"hi": 42, "id": 2}], JP[JP.hi].id, False, [2]),
        (
            [{"hello": 5, "id": 1}, {"hi": 42, "id": 2}],
            JP[JP.hi].id,
            True,
            [(JP[1].id, 2)],
        ),
        # cases: comparing to constant literal
        (
            {
                "product": [
                    {"current_price": 12.3, "avg_price": 10.5, "name": "kerfufle"},
                    {"current_price": 1.3, "avg_price": 9.5, "name": "acme"},
                ]
            },
            JP.product[JP.current_price < 10.1].name,
            False,
            ["acme"],
        ),
        (
            {
                "product": [
                    {"current_price": 12.3, "avg_price": 10.5, "name": "kerfufle"},
                    {"current_price": 1.3, "avg_price": 9.5, "name": "acme"},
                ]
            },
            JP.product[JP.current_price < 10.1].name,
            True,
            [(JP.product[1].name, "acme")],
        ),
        (
                {
                    "product": [
                        {"current_price": 12.3, "invalid": False, "name": "kerfufle"},
                        {"current_price": 1.3, "invalid": True, "name": "acme"},
                    ]
                },
                JP.product[JP.invalid == True].name,
                True,
                [(JP.product[1].name, "acme")],
        ),
        (
                {
                    "product": [
                        {"current_price": 12.3, "invalid": False, "name": "kerfufle"},
                        {"current_price": 1.3, "invalid": True, "name": "acme"},
                    ]
                },
                JP.product[JP.name == "kerfufle"].invalid,
                True,
                [(JP.product[0].invalid, False)],
        ),
        (
                {
                    "product": [
                        {"current_price": 12.3, "invalid": False, "name": None},
                        {"current_price": 1.3, "invalid": True, "name": "acme"},
                    ]
                },
                JP.product[JP.name == None].invalid,
                True,
                [(JP.product[0].invalid, False)],
        ),
        # cases: comparing list sub-item
        (
            [
                {"id": "a", "data": [0, 1]},
                {"id": "b", "data": [10, 11]},
                {"id": "c", "data": [20, 21]},
            ],
            JP[JP.data == [10, 11]].id,
            False,
            ["b"]
        ),
        (
            [
                {"id": "a", "data": [0, 1]},
                {"id": "b", "data": [10, 11]},
                {"id": "c", "data": [20, 21]},
            ],
            JP[JP.data == [10, 11]].id,
            True,
            [(JP[1].id, "b")]
        ),
        # cases: comparing dict sub-item
        (
[
                {"id": "a", "data": {"first": 0, "second": 1}},
                {"id": "b", "data": {"first": 10, "second": 11}},
                {"id": "c", "data": {"first": 20, "second": 21}},
            ],
            JP[JP.data == {"first": 10, "second": 11}].id,
            False,
            ["b"]
        ),
        (
            [
                {"id": "a", "data": {"first": 0, "second": 1}},
                {"id": "b", "data": {"first": 10, "second": 11}},
                {"id": "c", "data": {"first": 20, "second": 21}},
            ],
            JP[JP.data == {"first": 10, "second": 11}].id,
            True,
            [(JP[1].id, "b")]
        ),
        # cases: comparing two sub-items
        (
            {
                "product": [
                    {"current_price": 12.3, "avg_price": 10.5, "name": "kerfufle"},
                    {"current_price": 1.3, "avg_price": 9.5, "name": "acme"},
                ]
            },
            JP.product[JP.current_price > JP.avg_price].name,
            False,
            ["kerfufle"],
        ),
        (
            {
                "product": [
                    {"current_price": 12.3, "avg_price": 10.5, "name": "kerfufle"},
                    {"current_price": 1.3, "avg_price": 9.5, "name": "acme"},
                ]
            },
            JP.product[JP.current_price > JP.avg_price].name,
            True,
            [(JP.product[0].name, "kerfufle")],
        ),
    ],
)
def test_find_all_with_attribute_expressions(content, jp, with_path, expected_items):
    items = list(find_all(content, jp, with_path=with_path))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,with_path,expected_items",
    [
        ([0, 5, -5, 42], JP[JP == 0], True, [(JP[0], 0)]),
        ({"hello": [0, 5, -5, 42]}, JP.hello[JP == 0], True, [(JP.hello[0], 0)]),
        (
            [{"hi": 0}, {"hi": 5}, {"hi": -5}, {"hi": 42}],
            JP[JP.hi == 0],
            True,
            [(JP[0], {"hi": 0})],
        ),
        (
            {"hello": [{"hi": 0}, {"hi": 5}, {"hi": -5}, {"hi": 42}]},
            JP.hello[JP.hi == 0],
            True,
            [(JP.hello[0], {"hi": 0})],
        ),
        ([0, 5, -5, 42], JP[JP != 0], True, [(JP[1], 5), (JP[2], -5), (JP[3], 42)]),
        ([0, 5, -5, 42], JP[JP > 0], True, [(JP[1], 5), (JP[3], 42)]),
        ([0, 5, -5, 42], JP[JP >= 0], True, [(JP[0], 0), (JP[1], 5), (JP[3], 42)]),
        ([0, 5, -5, 42], JP[JP < 0], True, [(JP[2], -5)]),
        ([0, 5, -5, 42], JP[JP <= 0], True, [(JP[0], 0), (JP[2], -5)]),
    ],
)
def test_find_all_with_attribute_expressions_all_operators(
    content, jp, with_path, expected_items
):
    items = list(find_all(content, jp, with_path=with_path))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ([{"id2": 0}], JP[JP.id == 0], []),
        ([{"id": None}], JP[JP.id == 0], []),
        ([{"id": "not a number"}], JP[JP.id == 0], []),
    ],
)
def test_find_all_with_attribute_expressions_are_robust(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


###############################################################
#
# find_next()
#


@pytest.mark.parametrize(
    "content,jp,with_path,expected_value",
    [
        (
            [{"hello": 5}, {"hello": 42}],
            JP[:].hello,
            False,
            5,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[1].hello,
            False,
            42,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[:].hello,
            True,
            (JP[0].hello, 5),
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[1].hello,
            True,
            (JP[1].hello, 42),
        ),
    ],
)
def test_find_next(content, jp, with_path, expected_value):
    assert find_next(content, jp, with_path=with_path) == expected_value


@pytest.mark.parametrize(
    "content,jp,expected_error",
    [
        # case: simple, index error
        (
            [5, 42],
            JP[2],
            IndexError
        ),
        # case: simple, slice with empty list
        ([], JP[:], StopIteration),
        # case: simple, key error
        (
            {"hello": 5},
            JP.hi,
            KeyError
        ),
        # cases: mixed, longer paths
        ([], JP[:].hello, StopIteration),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hello,
            IndexError
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hi,
            IndexError
        ),
        ([], JP[1:].hello[:2].hi, StopIteration),
        ([], JP[:].hello[13].hi, StopIteration),
        ({}, JP.hello[1:].hi[:2], KeyError),
        ({}, JP.hello[13].hi[:], KeyError),
    ],
)
def test_find_next_nonexistent_without_default(content, jp, expected_error):
    for with_path in (True, False):
        with pytest.raises(expected_error):
            _ = find_next(content, jp, with_path=with_path)


@pytest.mark.parametrize(
    "content,jp",
    [
        # case: simple, ignoring index error
        (
            [5, 42],
            JP[2],
        ),
        # case: simple, slice with empty list, ignoring stop iteration error
        ([], JP[:]),
        # case: simple, ignoring key error
        (
            {"hello": 5},
            JP.hi,
        ),
        # cases: mixed, longer paths, ignoring errors
        ([], JP[:].hello),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hello,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hi,
        ),
        ([], JP[1:].hello[:2].hi),
        ([], JP[:].hello[13].hi),
        ({}, JP.hello[1:].hi[:2]),
        ({}, JP.hello[13].hi[:]),
    ],
)
def test_find_next_nonexistent_with_default(content, jp):
    defaults = (
        None,
        17,
        'a string',
        ('an ad-hoc object',)
    )
    with_paths = (True, False)

    for defaults, with_path in product(defaults, with_paths):
        results = find_next(content, jp, with_path=with_path, default=defaults)
        if with_path:
            assert isinstance(results, tuple)
            path, found = results
            assert path is None
            assert found == defaults
        else:
            found = results
            assert isinstance(found, type(defaults))
            assert found == defaults



from itertools import count

import pytest

from aweson import JP, find_all_duplicate, find_all_unique, with_values

###############################################################
#
# find_all_unique()
#


@pytest.mark.parametrize(
    "content,jp,with_path,expected_items",
    [
        ([1, 2, 1, 3, -22, 3], JP[:], False, [1, 2, 3, -22]),
        (
            [1, 2, 1, 3, -22, 3],
            JP[:],
            True,
            [(JP[0], 1), (JP[1], 2), (JP[3], 3), (JP[4], -22)],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            False,
            [1, 2, 3, -22],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            True,
            [(JP[0].hello, 1), (JP[1].hello, 2), (JP[3].hello, 3), (JP[4].hello, -22)],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:4].hello,
            False,
            [1, 2, 3],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:4].hello,
            True,
            [(JP[0].hello, 1), (JP[1].hello, 2), (JP[3].hello, 3)],
        ),
    ],
)
def test_find_all_unique(content, jp, with_path, expected_items):
    items = list(find_all_unique(content, jp, with_path=with_path))
    assert items == expected_items


###############################################################
#
# find_all_duplicate()
#


@pytest.mark.parametrize(
    "content,jp,with_path,expected_items",
    [
        ([1, 2, 1, 3, -22, 3], JP[:], False, [1, 3]),
        ([1, 2, 1, 3, -22, 3], JP[:], True, [(JP[2], 1), (JP[5], 3)]),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            False,
            [1, 3],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            True,
            [(JP[2].hello, 1), (JP[5].hello, 3)],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:3].hello,
            False,
            [1],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:3].hello,
            True,
            [(JP[2].hello, 1)],
        ),
    ],
)
def test_find_all_duplicates(content, jp, with_path, expected_items):
    items = list(find_all_duplicate(content, jp, with_path=with_path))
    assert items == expected_items


###############################################################
#
# with_values()
#


# mix case: several iterations on different level .foo[:]["h(ello|i)"][.bar == 0] ...
# TODO sub-hierarchy ???


def my_negator(n: int | None):
    """
    Silly negation for tests
    """
    if n is None:
        return -1 / 12
    if isinstance(n, int):
        return -n
    assert False


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        # - simple list
        ([11, 12, 13], JP[1], 137, [11, 137, 13]),
        (
            [11, 12, 13],
            JP[1],
            {"hi": 137},
            [11, {"hi": 137}, 13],
        ),
        # - list in the middle of path
        (
            {"hello": [{"hi": 11}, {"hi": 12}, {"hi": 13}]},
            JP.hello[1].hi,
            137,
            {"hello": [{"hi": 11}, {"hi": 137}, {"hi": 13}]},
        ),
        # - list is parent of leaf data
        ({"hello": [11, 12, 13]}, JP.hello[1], 137, {"hello": [11, 137, 13]}),
        # ITERATOR
        # - simple list
        ([11, 12, 13], JP[1], count(37, 1), [11, 37, 13]),
        (
            [11, 12, 13],
            JP[1],
            ({"hi": idx} for idx in count(37, 1)),
            [11, {"hi": 37}, 13],
        ),
        # - list in the middle of path
        (
            {"hello": [{"hi": 11}, {"hi": 12}, {"hi": 13}]},
            JP.hello[1].hi,
            count(37, 1),
            {"hello": [{"hi": 11}, {"hi": 37}, {"hi": 13}]},
        ),
        # - list is parent of leaf data
        ({"hello": [11, 12, 13]}, JP.hello[1], count(37, 1), {"hello": [11, 37, 13]}),
        # FUNCTION
        # - simple list
        ([11, 12, 13], JP[1], my_negator, [11, -12, 13]),
        # - list in the middle of path
        (
            {"hello": [{"hi": 11}, {"hi": 12}, {"hi": 13}]},
            JP.hello[1].hi,
            my_negator,
            {"hello": [{"hi": 11}, {"hi": -12}, {"hi": 13}]},
        ),
        # - list is parent of leaf data
        ({"hello": [11, 12, 13]}, JP.hello[1], my_negator, {"hello": [11, -12, 13]}),
    ],
)
def test_with_values_list_index(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        # - simple list
        ([11, 12, 13], JP[:2], 137, [137, 137, 13]),
        ([11, 12, 13], JP[:2], {"hi": 137}, [{"hi": 137}, {"hi": 137}, 13]),
        # - list in the middle of path
        (
            {"hello": [{"hi": 11}, {"hi": 12}, {"hi": 13}]},
            JP.hello[:2].hi,
            137,
            {"hello": [{"hi": 137}, {"hi": 137}, {"hi": 13}]},
        ),
        # - list is parent of leaf data
        ({"hello": [11, 12, 13]}, JP.hello[:2], 137, {"hello": [137, 137, 13]}),
        # ITERATOR
        # - simple list
        ([11, 12, 13], JP[:2], count(37, 1), [37, 38, 13]),
        (
            [11, 12, 13],
            JP[:2],
            ({"hi": idx} for idx in count(37, 1)),
            [{"hi": 37}, {"hi": 38}, 13],
        ),
        # - list in the middle of path
        (
            {"hello": [{"hi": 11}, {"hi": 12}, {"hi": 13}]},
            JP.hello[:2].hi,
            count(37, 1),
            {"hello": [{"hi": 37}, {"hi": 38}, {"hi": 13}]},
        ),
        # - list is parent of leaf data
        ({"hello": [11, 12, 13]}, JP.hello[:2], count(37, 1), {"hello": [37, 38, 13]}),
        # FUNCTION
        # - simple list
        ([11, 12, 13], JP[:2], my_negator, [-11, -12, 13]),
        # - list in the middle of path
        (
            {"hello": [{"hi": 11}, {"hi": 12}, {"hi": 13}]},
            JP.hello[:2].hi,
            my_negator,
            {"hello": [{"hi": -11}, {"hi": -12}, {"hi": 13}]},
        ),
        # - list is parent of leaf data
        ({"hello": [11, 12, 13]}, JP.hello[:2], my_negator, {"hello": [-11, -12, 13]}),
    ],
)
def test_with_values_list_slices(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        # - simple list
        #   - field existence
        (
            [
                {"foo": 10, "hola": 5},
                {"foo": 1, "hola": -42},
                {"foo": 2},
                {"foo": 3, "hola": None},
            ],
            JP[JP.hola].foo,
            137,
            [
                {"foo": 137, "hola": 5},
                {"foo": 137, "hola": -42},
                {"foo": 2},
                {"foo": 137, "hola": None},
            ],
        ),
        (
            [
                {"foo": 10, "hola": 5},
                {"foo": 1, "hola": -42},
                {"foo": 2},
                {"foo": 3, "hola": None},
            ],
            JP[JP.hola].foo,
            {"hi": 137},
            [
                {"foo": {"hi": 137}, "hola": 5},
                {"foo": {"hi": 137}, "hola": -42},
                {"foo": 2},
                {"foo": {"hi": 137}, "hola": None},
            ],
        ),
        #   - simple relational operator
        (
            [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}],
            JP[JP.hola < 0].foo,
            137,
            [{"foo": 10, "hola": 5}, {"foo": 137, "hola": -42}, {"foo": 2}],
        ),
        (
            [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}],
            JP[JP.hola < 0].foo,
            {"hi": 137},
            [{"foo": 10, "hola": 5}, {"foo": {"hi": 137}, "hola": -42}, {"foo": 2}],
        ),
        # - list in the middle of path
        #   - field existence
        (
            {
                "hello": [
                    {"foo": 10, "hola": 5},
                    {"foo": 1, "hola": -42},
                    {"foo": 2},
                    {"foo": 3, "hola": None},
                ]
            },
            JP.hello[JP.hola].foo,
            137,
            {
                "hello": [
                    {"foo": 137, "hola": 5},
                    {"foo": 137, "hola": -42},
                    {"foo": 2},
                    {"foo": 137, "hola": None},
                ]
            },
        ),
        #   - simple relational operator
        (
            {"hello": [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}]},
            JP.hello[JP.hola < 0].foo,
            137,
            {"hello": [{"foo": 10, "hola": 5}, {"foo": 137, "hola": -42}, {"foo": 2}]},
        ),
        # ITERATOR
        # - simple list: (a) field existence, (b) simple relational operator
        #   - field existence
        (
            [
                {"foo": 10, "hola": 5},
                {"foo": 1, "hola": -42},
                {"foo": 2},
                {"foo": 3, "hola": None},
            ],
            JP[JP.hola].foo,
            count(37, 1),
            [
                {"foo": 37, "hola": 5},
                {"foo": 38, "hola": -42},
                {"foo": 2},
                {"foo": 39, "hola": None},
            ],
        ),
        #   - simple relational operator
        (
            [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}],
            JP[JP.hola < 0].foo,
            count(37, 1),
            [{"foo": 10, "hola": 5}, {"foo": 37, "hola": -42}, {"foo": 2}],
        ),
        # - list in the middle of path
        #   - field existence
        (
            {
                "hello": [
                    {"foo": 10, "hola": 5},
                    {"foo": 1, "hola": -42},
                    {"foo": 2},
                    {"foo": 3, "hola": None},
                ]
            },
            JP.hello[JP.hola].foo,
            count(37, 1),
            {
                "hello": [
                    {"foo": 37, "hola": 5},
                    {"foo": 38, "hola": -42},
                    {"foo": 2},
                    {"foo": 39, "hola": None},
                ]
            },
        ),
        #   - simple relational operator
        (
            {"hello": [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}]},
            JP.hello[JP.hola < 0].foo,
            count(37, 1),
            {"hello": [{"foo": 10, "hola": 5}, {"foo": 37, "hola": -42}, {"foo": 2}]},
        ),
        # FUNCTION
        # - simple list
        #   - field existence
        (
            [
                {"foo": 10, "hola": 5},
                {"foo": 1, "hola": -42},
                {"foo": 2},
                {"foo": 3, "hola": None},
            ],
            JP[JP.hola].foo,
            my_negator,
            [
                {"foo": -10, "hola": 5},
                {"foo": -1, "hola": -42},
                {"foo": 2},
                {"foo": -3, "hola": None},
            ],
        ),
        #   - simple relational operator
        (
            [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}],
            JP[JP.hola < 0].foo,
            my_negator,
            [{"foo": 10, "hola": 5}, {"foo": -1, "hola": -42}, {"foo": 2}],
        ),
        # - list in the middle of path
        #   - field existence
        (
            {
                "hello": [
                    {"foo": 10, "hola": 5},
                    {"foo": 1, "hola": -42},
                    {"foo": 2},
                    {"foo": 3, "hola": None},
                ]
            },
            JP.hello[JP.hola].foo,
            my_negator,
            {
                "hello": [
                    {"foo": -10, "hola": 5},
                    {"foo": -1, "hola": -42},
                    {"foo": 2},
                    {"foo": -3, "hola": None},
                ]
            },
        ),
        #   - simple relational operator
        (
            {"hello": [{"foo": 10, "hola": 5}, {"foo": 1, "hola": -42}, {"foo": 2}]},
            JP.hello[JP.hola < 0].foo,
            my_negator,
            {"hello": [{"foo": 10, "hola": 5}, {"foo": -1, "hola": -42}, {"foo": 2}]},
        ),
    ],
)
def test_with_values_attribute_predicates(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        # - simple dict
        (
            {"hello": 5, "hi": 42},
            JP.hello,
            137,
            {"hello": 137, "hi": 42},
        ),
        # - dict in middle of the path
        (
            [{"hello": 5, "hi": 42}, {"hello": 29, "hi": 144}],
            JP[:].hello,
            137,
            [{"hello": 137, "hi": 42}, {"hello": 137, "hi": 144}],
        ),
        # - dict is parent of leaf data
        (
            {"foo": [{"hello": 5, "hi": 42}, {"hello": 29, "hi": 144}]},
            JP.foo[:].hello,
            137,
            {"foo": [{"hello": 137, "hi": 42}, {"hello": 137, "hi": 144}]},
        ),
        # ITERATOR
        # - simple dict
        (
            {"hello": 5, "hi": 42},
            JP.hello,
            count(37, 1),
            {"hello": 37, "hi": 42},
        ),
        # - dict in middle of the path
        (
            [{"hello": 5, "hi": 42}, {"hello": 29, "hi": 144}],
            JP[:].hello,
            count(37, 1),
            [{"hello": 37, "hi": 42}, {"hello": 38, "hi": 144}],
        ),
        # - dict is parent of leaf data
        (
            {"foo": [{"hello": 5, "hi": 42}, {"hello": 29, "hi": 144}]},
            JP.foo[:].hello,
            count(37, 1),
            {"foo": [{"hello": 37, "hi": 42}, {"hello": 38, "hi": 144}]},
        ),
        # FUNCTION
        # - simple dict
        (
            {"hello": 5, "hi": 42},
            JP.hello,
            my_negator,
            {"hello": -5, "hi": 42},
        ),
        # - dict in middle of the path
        (
            [{"hello": 5, "hi": 42}, {"hello": 29, "hi": 144}],
            JP[:].hello,
            my_negator,
            [{"hello": -5, "hi": 42}, {"hello": -29, "hi": 144}],
        ),
        # - dict is parent of leaf data
        (
            {"foo": [{"hello": 5, "hi": 42}, {"hello": 29, "hi": 144}]},
            JP.foo[:].hello,
            my_negator,
            {"foo": [{"hello": -5, "hi": 42}, {"hello": -29, "hi": 144}]},
        ),
    ],
)
def test_with_values_dict_key(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        # - simple dict
        (
            {"hello": 5, "helsinki": 13, "hi": 42},
            JP["hel.*"],
            137,
            {"hello": 137, "helsinki": 137, "hi": 42},
        ),
        # - dict in middle of the path
        (
            [
                {"hello": 5, "helsinki": 13, "hi": 42},
                {"hello": 29, "helsinki": 19, "hi": 144},
            ],
            JP[:]["hel.*"],
            137,
            [
                {"hello": 137, "helsinki": 137, "hi": 42},
                {"hello": 137, "helsinki": 137, "hi": 144},
            ],
        ),
        # - dict is parent of leaf data
        (
            {
                "foo": [
                    {"hello": 5, "helsinki": 13, "hi": 42},
                    {"hello": 29, "helsinki": 19, "hi": 144},
                ]
            },
            JP.foo[:]["hel.*"],
            137,
            {
                "foo": [
                    {"hello": 137, "helsinki": 137, "hi": 42},
                    {"hello": 137, "helsinki": 137, "hi": 144},
                ]
            },
        ),
        # ITERATOR
        # - simple dict
        (
            {"hello": 5, "helsinki": 13, "hi": 42},
            JP["hel.*"],
            count(37, 1),
            {"hello": 37, "helsinki": 38, "hi": 42},
        ),
        # - dict in middle of the path
        (
            [
                {"hello": 5, "helsinki": 13, "hi": 42},
                {"hello": 29, "helsinki": 19, "hi": 144},
            ],
            JP[:]["hel.*"],
            count(37, 1),
            [
                {"hello": 37, "helsinki": 38, "hi": 42},
                {"hello": 39, "helsinki": 40, "hi": 144},
            ],
        ),
        # - dict is parent of leaf data
        (
            {
                "foo": [
                    {"hello": 5, "helsinki": 13, "hi": 42},
                    {"hello": 29, "helsinki": 19, "hi": 144},
                ]
            },
            JP.foo[:]["hel.*"],
            count(37, 1),
            {
                "foo": [
                    {"hello": 37, "helsinki": 38, "hi": 42},
                    {"hello": 39, "helsinki": 40, "hi": 144},
                ]
            },
        ),
        # FUNCTION
        # - simple dict
        (
            {"hello": 5, "helsinki": 13, "hi": 42},
            JP["hel.*"],
            my_negator,
            {"hello": -5, "helsinki": -13, "hi": 42},
        ),
        # - dict in middle of the path
        (
            [
                {"hello": 5, "helsinki": 13, "hi": 42},
                {"hello": 29, "helsinki": 19, "hi": 144},
            ],
            JP[:]["hel.*"],
            my_negator,
            [
                {"hello": -5, "helsinki": -13, "hi": 42},
                {"hello": -29, "helsinki": -19, "hi": 144},
            ],
        ),
        # - dict is parent of leaf data
        (
            {
                "foo": [
                    {"hello": 5, "helsinki": 13, "hi": 42},
                    {"hello": 29, "helsinki": 19, "hi": 144},
                ]
            },
            JP.foo[:]["hel.*"],
            my_negator,
            {
                "foo": [
                    {"hello": -5, "helsinki": -13, "hi": 42},
                    {"hello": -29, "helsinki": -19, "hi": 144},
                ]
            },
        ),
    ],
)
def test_with_values_dict_regex(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        ([], JP[0], 137, [137]),
        ([5], JP[1], 137, [5, 137]),
        ({"hello": [5]}, JP.hello[1], 137, {"hello": [5, 137]}),
        # ITERATOR
        ([], JP[0], count(37, 1), [37]),
        ([5], JP[1], count(37, 1), [5, 37]),
        ({"hello": [5]}, JP.hello[1], count(37, 1), {"hello": [5, 37]}),
        # FUNCTION
        ([], JP[0], my_negator, [-1 / 12]),
        ([5], JP[1], my_negator, [5, -1 / 12]),
        ({"hello": [5]}, JP.hello[1], my_negator, {"hello": [5, -1 / 12]}),
    ],
)
def test_with_values_expanding_list(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        # VALUES
        ({}, JP.hello, 137, {"hello": 137}),
        ({"hi": 5}, JP.hello, 137, {"hi": 5, "hello": 137}),
        (
            [{"hi": 5}, {"hi": 5}],
            JP[:].hello,
            137,
            [{"hi": 5, "hello": 137}, {"hi": 5, "hello": 137}],
        ),
        # ITERATOR
        ({}, JP.hello, count(37, 1), {"hello": 37}),
        ({"hi": 5}, JP.hello, count(37, 1), {"hi": 5, "hello": 37}),
        (
            [{"hi": 5}, {"hi": 5}],
            JP[:].hello,
            count(37, 1),
            [{"hi": 5, "hello": 37}, {"hi": 5, "hello": 38}],
        ),
        # FUNCTION
        ({}, JP.hello, my_negator, {"hello": -1 / 12}),
        ({"hi": 5}, JP.hello, my_negator, {"hi": 5, "hello": -1 / 12}),
        (
            [{"hi": 5}, {"hi": 5}],
            JP[:].hello,
            my_negator,
            [{"hi": 5, "hello": -1 / 12}, {"hi": 5, "hello": -1 / 12}],
        ),
    ],
)
def test_with_values_adding_to_dicts(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        ({}, JP.hi, [137], {"hi": [137]}),
        ({}, JP.hi, {"hello": 137}, {"hi": {"hello": 137}}),
    ],
)
def test_with_values_non_primitive_values(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize(
    "content,jp",
    [
        ([0, 1], JP[3]),
        ({"hello": [0, 1]}, JP.hello[3]),
    ],
)
def test_with_values_rejects_gaps_in_lists(content, jp):
    with pytest.raises(ValueError, match="Cannot append at"):
        _ = with_values(content, jp, 0)


@pytest.mark.parametrize(
    "content,jp,values",
    [
        ([0, 1, 2], JP[:], iter([1, 2])),
        ([{"hello": 5}, {"hello": 42}, {"hello": 137}], JP[:].id, iter([1, 2])),
        ([{"hello": 5}, {"hello": 42}, {"hello": 137}], JP[JP.hello].id, iter([1, 2])),
    ],
)
def test_with_values_insufficient_iterated_input_rejected(content, jp, values):
    with pytest.raises(ValueError, match="Iterator size falls short of expectations"):
        _ = with_values(content, jp, values)


def test_with_values_demo_id_generation():
    content = [
        {"hello": 5},
        {"hello": 42},
        {"hello": 137},
    ]
    content_with_ids = with_values(content, JP[:].id, count(0, 1))
    assert content_with_ids == [
        {"hello": 5, "id": 0},
        {"hello": 42, "id": 1},
        {"hello": 137, "id": 2},
    ]


@pytest.mark.parametrize(
    "content,jp,values,expected_content",
    [
        # specific path
        (
            {
                "hello": [
                    {"hi": "world", "details": [5, 10]},
                    {"hi": "maailma", "details": [7, 20]},
                    {"hi": "mundo", "details": [5, 30]},
                ],
                "irrelevant": 3.14,
            },
            JP.hello[1].details[1],
            137,
            {
                "hello": [
                    {"hi": "world", "details": [5, 10]},
                    {"hi": "maailma", "details": [7, 137]},
                    {"hi": "mundo", "details": [5, 30]},
                ],
                "irrelevant": 3.14,
            },
        ),
        # over multiple slices
        (
            {
                "hello": [
                    {"hi": "world", "details": [5, 10]},
                    {"hi": "maailma", "details": [7, 20]},
                    {"hi": "mundo", "details": [5, 30]},
                ],
                "irrelevant": 3.14,
            },
            JP.hello[:].details[:],
            my_negator,
            {
                "hello": [
                    {"hi": "world", "details": [-5, -10]},
                    {"hi": "maailma", "details": [-7, -20]},
                    {"hi": "mundo", "details": [-5, -30]},
                ],
                "irrelevant": 3.14,
            },
        ),
        # over multiple slices (actual slices)
        (
            {
                "hello": [
                    {"hi": "world", "details": [5, 10]},
                    {"hi": "maailma", "details": [7, 20]},
                    {"hi": "mundo", "details": [5, 30]},
                ],
                "irrelevant": 3.14,
            },
            JP.hello[1:].details[1:],
            my_negator,
            {
                "hello": [
                    {"hi": "world", "details": [5, 10]},
                    {"hi": "maailma", "details": [7, -20]},
                    {"hi": "mundo", "details": [5, -30]},
                ],
                "irrelevant": 3.14,
            },
        ),
        # mkdir -p
        ({}, JP.hello.hi.moi, 137, {"hello": {"hi": {"moi": 137}}}),
        # mkdir -p #2
        ({}, JP["hello"].hi["moi"], 137, {"hello": {"hi": {"moi": 137}}}),
        # mkdir -p with lists
        (
            [],
            JP[0][0][0],
            137,
            [
                [
                    [
                        137,
                    ],
                ],
            ],
        ),
    ],
)
def test_with_values_misc_cases(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


# TODO test case: overwrite entire dictionaries (with a function)

counter = count(137, 1)


@pytest.mark.parametrize(
    "content, jp, values, expected_content",
    [
        (
            [
                {"product": "apple", "price": -3},
                {"product": "pear", "price": 3},
                {"product": "peach", "price": 0},
            ],
            JP[:],
            (
                lambda r: (
                    r
                    if r["price"] > 0
                    else (
                        r | {"zero": True}
                        if r["price"] == 0
                        else r | {"error": "Negative price"}
                    )
                )
            ),
            [
                {"product": "apple", "price": -3, "error": "Negative price"},
                {"product": "pear", "price": 3},
                {"product": "peach", "price": 0, "zero": True},
            ],
        ),
        (
            [
                {"product": "apple", "price": -3},
                {"product": "pear", "price": 3},
                {"product": "peach", "price": 0},
            ],
            JP[:](JP.error),
            (
                lambda r: (
                    (None,)
                    if r["price"] > 0
                    else ("Zero price",) if r["price"] == 0 else ("Negative price",)
                )
            ),
            [
                {"product": "apple", "price": -3, "error": "Negative price"},
                {"product": "pear", "price": 3, "error": None},
                {"product": "peach", "price": 0, "error": "Zero price"},
            ],
        ),
        (
            [
                {"product": "apple", "price": -3},
                {"product": "pear", "price": 3},
                {"product": "peach", "price": 0},
            ],
            JP[JP.price <= 0](JP.error),
            (
                lambda r: (
                    (None,)
                    if r["price"] > 0
                    else ("Zero price",) if r["price"] == 0 else ("Negative price",)
                )
            ),
            [
                {"product": "apple", "price": -3, "error": "Negative price"},
                {"product": "pear", "price": 3},
                {"product": "peach", "price": 0, "error": "Zero price"},
            ],
        ),
        (
            [{}, {}, {}],
            JP[:](JP.product, JP.price, JP.id),
            zip(iter(["apple", "pear", "peach"]), iter([1.2, 2.3, 3.4]), count(0, 1)),
            [
                {"product": "apple", "price": 1.2, "id": 0},
                {"product": "pear", "price": 2.3, "id": 1},
                {"product": "peach", "price": 3.4, "id": 2},
            ],
        ),
        (
            [{"price": 1.2}, {"price": 2.3}, {"price": 123.4}],
            JP[:](JP.id, JP.verdict),
            lambda d: (
                (next(counter), "expensive")
                if d["price"] > 100
                else (next(counter), "OK")
            ),
            [
                {"price": 1.2, "id": 137, "verdict": "OK"},
                {"price": 2.3, "id": 138, "verdict": "OK"},
                {"price": 123.4, "id": 139, "verdict": "expensive"},
            ],
        ),
    ],
)
def test_with_values_composite(content, jp, values, expected_content):
    new_content = with_values(content, jp, values)
    assert new_content == expected_content


@pytest.mark.parametrize("jp_sub_item", [JP.too.long.path, JP[".*"], JP[0], JP[:]])
def test_with_values_composite_complex_sub_items_rejected(jp_sub_item):
    with pytest.raises(ValueError):
        _ = with_values([{}], JP[:](jp_sub_item), 5)
